From mathcomp Require Import ssreflect.all_ssreflect.
Require Import axioms.
Set Implicit Arguments.
Unset Strict Implicit.

(** ** Composition *)

Notation "f >> g" := (funcomp tt g f) (*fun x => g (f x)*) (at level 50) : subst_scope.
Open Scope subst_scope.

(** ** Finite Types *)

Fixpoint fin (n : nat) : Type :=
  match n with
  | 0 => False
  | S m => option (fin m)
  end.

Definition scons {X : Type} {n : nat} (x : X) (f : fin n -> X) (m : fin (S n)) : X :=
  match m with
  | None => x
  | Some i => f i
  end.
Notation "x .: f" := (@scons _ _ x f) (at level 55) : subst_scope.

Definition ren (m n : nat) : Type := fin m -> fin n.

Set Primitive Projections.
Structure iren (m n : nat) := IRen {
  iren_fun :> fin m -> fin n;
  iren_inj : injective iren_fun
}.
Arguments IRen {m n} iren_fun iren_inj.

Lemma iren_eq {m n} (f g : iren m n) :
  (forall x, f x = g x) -> f = g.
Proof.
  move: f g => -[f fP] [g gP] /= H. have E: f = g by fext.
  destruct E. f_equal. exact: pi.
Qed.

Definition bound {n : nat} : fin n.+1 := None.
Definition null {T} (i : fin 0) : T := match i with end.

Definition rcomp {m n k : nat} (f : iren m n) (g : iren n k) : iren m k :=
  @IRen m k (f >> g) (fun x y e => iren_inj (iren_inj e)).
Notation "f >>> g" := (rcomp f g) (at level 50) : subst_scope.

Definition idren {n : nat} : iren n n :=
  @IRen n n id (fun x y e => e).

Definition shift {n : nat} : iren n n.+1 :=
  @IRen n n.+1 Some (fun x y e => f_equal (fun o => if o is Some z then z else x) e).
Arguments shift {n} : simpl never.

Definition up {m n : nat} (f : iren m n) : iren m.+1 n.+1.
  refine (@IRen m.+1 n.+1 (bound .: f >>> shift) _).
  abstract (by move=>/=[x|] [y|] //= [/iren_inj->]).
Defined.

Lemma shift_up {m n} (f : iren m n) :
  shift >>> up f = f >>> shift.
Proof. exact: iren_eq. Qed.

Lemma scons_eta {T} {n : nat} (f : fin n.+1 -> T) :
  f bound .: shift >> f = f.
Proof. by fext=>/=;case. Qed.

Lemma scons_eta_id {n : nat} : bound .: shift = id :> (fin n.+1 -> fin n.+1).
Proof. by fext=>/=;case. Qed.

Lemma scons_comp {T1 T2} {n : nat} (x : T1) (f : fin n -> T1) (g : T1 -> T2) :
  (x .: f) >> g = (g x) .: f >> g.
Proof. by fext=>/=;case. Qed.

Ltac fsimpl :=
  repeat match goal with
    | [|- context[id >> ?f]] => change (id >> f) with f
    | [|- context[?f >> id]] => change (f >> id) with f
    | [|- context[(?f >> ?g) >> ?h]] =>
        change ((f >> g) >> h) with (f >> (g >> h))
    | [|- context[idren >>> ?f]] => change (idren >>> f) with f
    | [|- context[?f >>> idren]] => change (f >>> idren) with f
    | [|- context[(?f >>> ?g) >>> ?h]] =>
        change ((f >>> g) >>> h) with (f >>> (g >>> h))
    | [|- context[?f >> (?x .: ?g)]] =>
        change (f >> (x .: g)) with g
    (*| [|- context[shift >> (?x1 .: ?xr)]] =>
        change (shift >> (x1 .: xr)) with xr*)
    (*| [|- context[Some >> (?x1 .: ?xr)]] =>
        change (Some >> (x1 .: xr)) with xr*)
    | [|- context[?x2 .: shift >> ?f]] =>
        change x2 with (f bound); rewrite (@scons_eta _ _ f)
    | [|- context[?f bound .: ?g]] =>
        change g with (shift >> f); rewrite scons_eta
    (*| [|- context[?x2 .: Some >> ?f]] =>
        change x2 with (f bound); rewrite (@scons_eta _ _ f)*)
    | _ => progress (rewrite ?scons_comp ?scons_eta_id)
  end.
